/*
 * NATDet - NAT Detection tool
 * ---------------------------
 * Copyright (c) 2003 - 2006 Marcin Ulikowski <elceef@itsec.pl>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifndef WIN32
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <pwd.h>
#include <ncurses.h>
#else
#pragma comment (lib, "WSOCK32.lib")
#pragma comment (lib, "wpcap.lib")
#endif
#include <pcap.h>
#include <signal.h>
#include <time.h>

#include "config.h"
#include "tcpip.h"
#include "struct.h"


static struct nat_cache nat[NAT_CACHE]; /* NAT cache entries */
static struct fp_cache fp[MAX_SIGS];    /* OS fingerprints */
static struct config cfg;               /* configuration */

static char *iface, *rule;              /* device and BPF rule */
static u_char fhlen;                    /* frame header length */

static u_short cpos, sigs;

struct pcap_pkthdr header;
struct pcap_stat pstat;

u_int pcount = 0;


static void bug(char *format, ...) {
  char buffer[256];
  va_list args;

  va_start(args, format);
  vsnprintf(buffer, sizeof(buffer) - 1, format, args);
  fprintf(stderr, "[!] %s", buffer);
  va_end(args);

  exit(1);
}


/* find out frame header length */
static void set_fhlen(u_int dlink_type) {
  switch(dlink_type) {

    case DLT_RAW:        break;

    case DLT_SLIP:       fhlen = 16;
                         break;

    case DLT_EN10MB:     fhlen = 14;
                         break;

#ifdef DLT_LOOP
    case DLT_LOOP:
#endif
    case DLT_NULL:
    case DLT_PPP:        fhlen = 4;
                         break;

    case DLT_IEEE802:    fhlen = 22;
                         break;

#ifdef DLT_PPP_ETHER
    case DLT_PPP_ETHER:  fhlen = 8;
                         break;
#endif

#ifdef DLT_PFLOG
    case DLT_PFLOG:      fhlen = 28;
                         break;
#endif

#ifdef DLT_LINUX_SLL
    case DLT_LINUX_SLL:  fhlen = 16;
                         break;
#endif

#ifdef DLT_SLIP_BSDOS
    case DLT_SLIP_BSDOS: fhlen = 24;
                         break;
#endif

#ifdef DLT_PPP_BSDOS
    case DLT_PPP_BSDOS:  fhlen = 24;
                         break;
#endif

#ifdef DLT_PPP_SERIAL
    case DLT_PPP_SERIAL: break;
#endif

    default:             fprintf(stderr, "[!]: unknown datalink type (%u)!\n", dlink_type);
                         break;
  }

}


static char *osfp(u_short window, u_char ttl, u_short df, u_short plen, u_short mss) {
  u_short sigwss = 0, i;

  for (i = 0; i < sigs; i++) {

    if (tolower(fp[i].win[0]) == 'm') {
      if (fp[i].win[1] == '*')
        sigwss = atoi(fp[i].win+2) * mss;
      if (fp[i].win[1] == '+')
        sigwss = atoi(fp[i].win+2) + mss;
    } else {
      sigwss = atoi(fp[i].win);
    }

    if (sigwss == window) {

      if (fp[i].len == plen || !fp[i].len) {

        if ((df == IP_DF && fp[i].df == 1) || (!df && !fp[i].df)) {

          if (fp[i].ttl == 32 && ttl <= 32) {  /* (0; 32> */
            return fp[i].os;
          } else
          if (fp[i].ttl == 64 && ttl <= 64 && ttl > 32) {  /* (32; 64> */
            return fp[i].os;
          } else
          if (fp[i].ttl == 128 && ttl <= 128 && ttl > 64) {  /* (64; 128> */
            return fp[i].os;
          } else
          if (fp[i].ttl == 255 && ttl > 128) {  /* (128; 255> */
            return fp[i].os;
          }

        }
      }
    }
  }

  return "Unknown";
}


static char *showtime(void) {
  static char buff[20];
  char *pbuff = buff;
  time_t i = time(0);
  struct tm *t;

  t = localtime((void *)&i);
  snprintf(buff, sizeof(buff), "%02u-%02u-%u %02u:%02u:%02u", t->tm_mday,
           t->tm_mon+1, t->tm_year+1900, t->tm_hour, t->tm_min, t->tm_sec);

  return pbuff;
}


static char *resolv(u_int ip) {
  struct hostent *addr;
  u_char *iaddr;
  static char dotip[16];

  iaddr = (u_char *)&ip;
  snprintf(dotip, sizeof(dotip), "%u.%u.%u.%u",
           iaddr[0], iaddr[1], iaddr[2], iaddr[3]);
  addr = gethostbyaddr((char *)&ip, 4, AF_INET);
  if (!addr) return dotip;

  return addr->h_name;
}


static char *long2ip(u_int ip) {
  u_char *iaddr;
  static char dotip[16];

  iaddr = (u_char *)&ip;
  snprintf(dotip, sizeof(dotip), "%u.%u.%u.%u",
           iaddr[0], iaddr[1], iaddr[2], iaddr[3]);
  return dotip;
}


static void load_signatures(void) {
  FILE *osprints;
  FILE *test_osprints;

  u_int osdf, osttl, osplen,
        count = 0;
  u_int test_osdf, test_osttl, test_osplen,
        test_count = 0;

  char line[SIG_LEN], osname[SIG_LEN], oswin[SIG_LEN];
  char test_line[SIG_LEN], test_oswin[SIG_LEN];
  char *l;

  if (!(osprints = fopen(FP_DB, "r")))
    if (!(osprints = fopen(FP_DIR"/"FP_DB, "r")))
      bug("unable to open '%s'\n", FP_DB);

  while (count++ <= MAX_SIGS) {
 
    l = fgets(line, SIG_LEN, osprints);

    if (feof(osprints)) break;

    if (line[0] == '#' || line[0] == '\n')
      continue;

    if (sscanf(line, "%[0-9*+M]:%u:%u:%u:%[^:]", oswin, &osttl, &osdf, &osplen, osname) != 5) {
      bug("syntax error on line %u in '%s'\n", count, FP_DB);
    }

    /* test signature */
    if (!(test_osprints = fopen(FP_DB, "r")))
      if (!(test_osprints = fopen(FP_DIR"/"FP_DB, "r")))
        bug("unable to open '%s'\n", FP_DB);

    test_count = 0;
    while (test_count++ <= MAX_SIGS) {

      l = fgets(test_line, SIG_LEN, test_osprints);

      if (feof(test_osprints)) break;

      if (test_line[0] == '#' || test_line[0] == '\n')
        continue;

      if (sscanf(test_line, "%[0-9*+M]:%u:%u:%u:", test_oswin, &test_osttl, &test_osdf, &test_osplen) != 4)
        bug("syntax error on line %u in '%s'\n", test_count, FP_DB);

      if (test_oswin  == oswin  &&
          test_osttl  == osttl  &&
          test_osdf   == osdf   &&
          test_osplen == osplen &&
          test_count  != count) {
        bug("double signatures on lines %u and %u in '%s'\n", count, test_count, FP_DB);
      }

    }

    fclose(test_osprints);
    /* signature tested */

    if ((tolower(oswin[0]) != 'm' && atoi(oswin) > 65535) ||
        (tolower(oswin[0]) == 'm' && atoi(oswin+2) == 0))
      bug("invalid WSS on line %u in '%s'\n", count, FP_DB);

    if (osttl % 32 && osttl != 255)
      bug("invalid TTL %d on line %u in '%s'\n", osttl, count, FP_DB);

    if (osdf > 1)
      bug("invalid DF (%d) on line %u in '%s'\n", osdf, count, FP_DB);

    if (osplen > 64 || osplen % 4)
      bug("invalid packet length on line %u in '%s'\n", count, FP_DB);

    if (!(strlen(osname) - 1))
      bug("empty OS description on line %u in '%s'\n", count, FP_DB);

    if (sigs > MAX_SIGS) break;

    osname[strlen(osname) - 1] = '\0';

    fp[sigs].win = strdup(oswin);
    fp[sigs].ttl = osttl;
    fp[sigs].len = osplen;
    fp[sigs].os  = strdup(osname);
    if (osdf) fp[sigs].df = 1;
    else fp[sigs].df = 0;

    sigs++;
  }

  fclose(osprints);
}


static void watch(u_char *args, struct pcap_pkthdr *header, u_char *packet) {
  u_short ethtype;
  u_char *ethp = (u_char *)(packet + 12);
  memcpy(&ethtype, ethp, 2);
#if BYTE_ORDER == LITTLE_ENDIAN
  ethtype = ntohs(ethtype);
#endif
  if (ethtype == 0x8100) fhlen += 4; // 802.1Q VLAN

  struct iphdr  *ip  = (struct iphdr *)(packet + fhlen);
  struct tcphdr *tcp = (struct tcphdr *)(packet + fhlen + (ip->ihl * 4));

  if (ethtype == 0x8100) fhlen -= 4; // 802.1Q VLAN

  u_char *optp, *optp_end;
#ifndef WIN32
  char host[30];
#endif

  static char ttl_diff, ttl1, ttl2,
              osgenre[16];

  u_char fcount = 0,
         factor = 0;

  u_short oldprob, tmpprob, i, j, k;

  u_int timestamp = 0;
  u_short mss = 0;

  u_long cur_time = 0;

#if BYTE_ORDER == LITTLE_ENDIAN
  ip->tot_len  = htons(ip->tot_len);
  ip->frag_off = htons(ip->frag_off);

  tcp->window  = htons(tcp->window);
  tcp->source  = htons(tcp->source);
#endif

  /* just in case ... */
  if (ip->ihl < 5 && ip->version != 0x04) {
    printf("[i] Oops! Seems we've received an invalid TCP/IP packet :)\n");
    goto end;
  }

  optp  = (u_char *)(tcp + 1);
  optp_end = optp + ((tcp->doff) << 2) - sizeof(struct tcphdr);

  while (optp < optp_end) {
    switch (*(optp++)) {

      case TCP_TSTAMP:
        memcpy(&timestamp, optp+1, 4);
        optp += 9;
#if BYTE_ORDER == LITTLE_ENDIAN
        timestamp = ntohl(timestamp);
#endif
      break;

      case TCP_MSS:
        memcpy(&mss, optp+1, 2);
        optp += 3;
#if BYTE_ORDER == LITTLE_ENDIAN
        mss = ntohs(mss);
#endif
        break;

      case TCP_NOP:
        break;

      case TCP_SACKOK:
        optp++;
        break;

      case TCP_WSCALE:
        optp += 2;
        break;

      case TCP_EOL:
        goto tcpopt_end;
        break;

      default:
        break;

    }
  }

tcpopt_end: ;

  strncpy(osgenre, osfp(tcp->window, ip->ttl, ip->frag_off, ip->tot_len, mss),
          sizeof(osgenre));

#ifdef _DEBUG
#ifndef WIN32
  if (!cfg.iactive)
#endif
  printf("%s.%u: ttl %u win %u len %u %smss %u tstamp %u %s\n",
         long2ip(ip->saddr), tcp->source, ip->ttl, tcp->window,
         ip->tot_len, (ip->frag_off == IP_DF) ? "DF " : "", mss,
         timestamp, osgenre);
#endif

  cur_time = time(0);

  if (ip->ttl > 0 && ip->ttl <= 32) {
    ttl1 = 32 - ip->ttl;

  } else if (ip->ttl > 32 && ip->ttl <= 64) {
    ttl1 = 64 - ip->ttl;

  } else if (ip->ttl > 64 && ip->ttl <= 128) {
    ttl1 = 128 - ip->ttl;

  } else if (ip->ttl > 128)
    ttl1 = 255 - ip->ttl;

  if (ip->saddr == ip->daddr) goto end;

  for (i = 0; i < cpos; i++) {

    if (nat[i].ip != ip->saddr && nat[i].ip && nat[i].prob && nat[i].prob < 100) {
      nat[i].prob -= ((cur_time-nat[i].time)/LIFE_TIME*100 <= nat[i].prob) ? (cur_time-nat[i].time)/LIFE_TIME*100 : 0;
    }

    if (nat[i].ip != ip->saddr || nat[i].ip == 0)
      continue;

    if (nat[i].ttl > 0 && nat[i].ttl <= 32) {
      ttl2 = 32 - nat[i].ttl;

    } else if (nat[i].ttl > 32 && nat[i].ttl <= 64) {
      ttl2 = 64 - nat[i].ttl;

    } else if (nat[i].ttl > 64 && nat[i].ttl <= 128) {
      ttl2 = 128 - nat[i].ttl;

    } else if (nat[i].ttl > 128)
      ttl2 = 255 - nat[i].ttl;

    ttl_diff = abs(ttl1 - ttl2);

    oldprob = nat[i].prob;
    tmpprob = 0;

    /* do not count "Unknown" systems */
    if (strcmp(osgenre, "Unknown")) {
      for (k = 0; k < OS_LIMIT; k++) {
        if (!strcmp(nat[i].osl[k], osgenre))
          break;
        /* add another OS to the counter */
        if (!strlen(nat[i].osl[k])) {
          strncpy(nat[i].osl[k], osgenre, sizeof(nat[i].osl[k]));
          nat[i].oscnt++;
          break;
        }
      }
    }

    if (ip->tot_len >= 40 && ip->tot_len <= 64) {
      for (k = 0; k < 8; k++) {
        if (nat[i].synlen[k] == ip->tot_len)
          break;
        if (!nat[i].synlen[k]) {
          nat[i].synlen[k] = ip->tot_len;
          nat[i].synlencnt++;
          break;
        }
      }
    }

    /* if cache entry is too old, rewrite it with new data */
    if (cur_time - nat[i].time >= LIFE_TIME) {
      nat[i].time   = cur_time;
      nat[i].ttl    = ip->ttl;
      nat[i].tstamp = timestamp;
      nat[i].prob   = 0;

      nat[i].oscnt  = 1;
      strncpy(nat[i].osgenre, osgenre, sizeof(nat[i].osgenre));

      bzero(nat[i].osl, sizeof(nat[i].osl));
      strncpy(nat[i].osl[0], osgenre, sizeof(nat[i].osl[0]));

      nat[i].synlencnt = 1;
      bzero(nat[i].synlen, sizeof(nat[i].synlen));
      nat[i].synlen[0] = ip->tot_len;

      goto end;
    }


    /* different overall SYN packet lengths */
    if (cfg.factor & SYNLEN &&
        nat[i].synlencnt > 1) {
      tmpprob += 12;
      factor |= SYNLEN;
      fcount++;
    }

    /* if osgenres differs and both are known, use this factor */
    if (cfg.factor & OSGENRE &&
        strcmp(nat[i].osgenre, osgenre) && strcmp(osgenre, "Unknown") && strcmp(nat[i].osgenre, "Unknown")) {
      if (nat[i].oscnt >= MIN_OSCOUNT) {
        tmpprob += 16;
      } else {
        tmpprob += nat[i].oscnt * 4;
      }
      factor |= OSGENRE;
      fcount++;
    }

    /* 1. if timestamp is lower than cached, use this factor
       2. we assume that timestamp increases by 100 each second
          so we calculate what value it should be and if it's bigger than
          the current timestamp, we use this factor */
    if (cfg.factor & TSTAMP &&
        ((timestamp > nat[i].tstamp && (timestamp > (cur_time - nat[i].time) * 100 + nat[i].tstamp + TSTAMP_DIFF)) ||
        (timestamp != nat[i].tstamp && (timestamp + TSTAMP_DIFF < nat[i].tstamp || (!nat[i].tstamp && timestamp) || (!timestamp && nat[i].tstamp))))) {

      tmpprob += 4;

      /* don't forget to refresh tcp-timestamp if osgenre is the same */
      if (timestamp && !strcmp(nat[i].osgenre, osgenre) && nat[i].ttl == ip->ttl) {
        nat[i].tstamp = timestamp;
        nat[i].time = cur_time;
      }

      factor |= TSTAMP;
      fcount++;
    }

    /* TTL difference? use this factor! */
    if (cfg.factor & TTL &&
        ttl_diff > 0 && ttl_diff <= HOP_DIST) {
      tmpprob += 6;
      factor |= TTL;
      fcount++;
    }

    /* factors below are less reliable */

    /* fails with source-port randomization */
    if (cfg.factor & SRCPORT &&
        tcp->source > 49152) {
      tmpprob += 1;
      factor |= SRCPORT;
      fcount++;
    }

    /* MSS tweaking - sometimes useful */
    if (cfg.factor & MSSWSS && mss) {
      if ((tcp->window % mss) && !(tcp->window % 1460)) {
        tmpprob += 1;
        factor |= MSSWSS;
        fcount++;
      }
    }

    /* refresh record time */
    if (!strcmp(nat[i].osgenre, osgenre) && nat[i].ttl == ip->ttl && !(factor & TSTAMP))
      nat[i].time = cur_time;


    if (fcount && nat[i].prob < 100) {
      tmpprob = tmpprob * 100 / (1+1+4+6+12+16);
      /* 100% sure conditions */
      if ( (factor & TTL && fcount >= 2) ||
           (factor & OSGENRE && factor & TSTAMP && nat[i].oscnt >=2) ||
           nat[i].oscnt >= MIN_OSCOUNT ||
           nat[i].synlencnt >= MIN_SYNLEN ) {
        nat[i].prob = 100;
      } else {
        /* calculate probability for the host */
        nat[i].prob = (tmpprob * (LIFE_TIME - (cur_time - nat[i].time)) / LIFE_TIME + oldprob) / 2;
        if (nat[i].prob > nat[i].maxprob) nat[i].maxprob = nat[i].prob;
      }
    }

#ifndef WIN32
    /* For those who just hate reading logs :)) */
    if (cfg.iactive) {
      int row, col, barlen, lcount;

      erase();

      getmaxyx(stdscr, row, col);

      if (col - 40 > 0 && col - 40 < 256) {
        barlen = col - 40;
      } else {
        barlen = 40;
      }

      attron(A_REVERSE);

      for (j = 0; j <= col; j++) {
        mvprintw(0, j, " ");
      }
      mvprintw(0, 0, "NATDet "VERSION);
      mvprintw(0, col-strlen(iface)-29, "device:%s | %s", iface, showtime());

      attroff(A_REVERSE);

      move(1, 32+(cfg.minprob*barlen/100));addch('v');printw(" %u%%", cfg.minprob);

      move(2, 0);

      lcount = 0;

      for (k = 0; k <= cpos; k++) {

        if (nat[k].prob >= cfg.minprob && lcount < row-2) {
          bzero(host, sizeof(host));

          snprintf(host, 30, "%s", (!cfg.notresolv) ? nat[k].hostname : long2ip(nat[k].ip));

          if (nat[k].prob == 100) attron(A_BOLD);
          printw(" %-30s |", host);
          for (j = 0; j < nat[k].prob*barlen/100; j++) addch(ACS_CKBOARD);
          if (nat[k].maxprob >= nat[k].prob) {
            for (j = 0; j < (nat[k].maxprob - nat[k].prob)*barlen/100; j++) addch(' ');
            for (j = 0; j < barlen-(nat[k].prob*barlen/100)-(nat[k].maxprob - nat[k].prob)*barlen/100; j++) addch(ACS_BULLET);
          } else {
            for (j = 0; j < barlen-(nat[k].prob*barlen/100); j++) addch(ACS_BULLET);
          }
          printw("| %3u%%\n", nat[k].prob);
          if (nat[k].prob == 100) attroff(A_BOLD);

          lcount++;
        }
      } /* for() */

      /* when screen is crowded */
      if (lcount >= row-2 && cfg.minprob < 100)
        cfg.minprob++;

      if (cfg.verbose) {
        attron(A_REVERSE);

        bzero(host, sizeof(host));

        for (j = 0; j <= col; j++) {
          mvprintw(row-1, j, " ");
        }

        mvprintw(row-1, 0, "#%u %s: ttl %u win %u len %u %smss %u tstamp %u %s",
                 pcount++, long2ip(ip->saddr), ip->ttl, tcp->window, ip->tot_len,
                 (ip->frag_off == IP_DF) ? "DF " : "", mss, timestamp, osgenre);

        attroff(A_REVERSE);
      }

      refresh();

    } else
#endif
    /* usual output */
    if (fcount >= MIN_FACTORS   &&
        nat[i].prob >= cfg.minprob &&
        oldprob < nat[i].prob   &&
        oldprob + REP_THRESHOLD <= nat[i].prob) {


      printf("[*] %s: NAT at %s for %u system(s) [%u%%]\n",
             showtime(), (!cfg.notresolv) ? resolv(ip->saddr) : long2ip(ip->saddr),
             nat[i].oscnt, nat[i].prob);

      if (cfg.verbose) {
        if (nat[i].ttl > ip->ttl) {
          if (nat[i].tstamp > 360000)
            printf("    Estimated gateway uptime: %uday(s) %uhour(s)\n",
                   nat[i].tstamp/360000/24, nat[i].tstamp/360000%24);
        } else {
          if (timestamp > 360000)
            printf("    Estimated gateway uptime: %uday(s) %uhour(s)\n",
                   timestamp/360000/24, timestamp/360000%24);
        }

        if (nat[i].oscnt > 1) {
          printf("    Recognized systems: ");
          for (j = 0; j < nat[i].oscnt; j++) {
            printf("%s ", nat[i].osl[j]);
          }
          putchar('\n');
        }

      }

      if (cfg.verbose == 2) {
        printf("    Used factors: ");
        if (factor & OSGENRE) printf("OSGENRE ");
        if (factor & TSTAMP)  printf("TSTAMP ");
        if (factor & TTL)     printf("TTL ");
        if (factor & SRCPORT) printf("SRCPORT ");
        if (factor & MSSWSS)  printf("MSSWSS ");
        if (factor & SYNLEN)  printf("SYNLEN ");
        putchar('\n');

        printf("    #%u: %s, %u, %u (%lus) -> %s, %u, %u\n",
               i, nat[i].osgenre, nat[i].ttl, nat[i].tstamp,
               cur_time-nat[i].time, osgenre, ip->ttl, timestamp);
      }

    }

    goto end;

  } /* loop */

  if (cpos == NAT_CACHE) {
    cpos = 0;
    if (!cfg.iactive)
      printf("[i] Cache memory is full. Overwriting existing cache entries.\n"
             "    Use filter expressions to avoid this in future.\n");
  }

  /* so, we have new packet to remember */
  nat[cpos].ttl     = ip->ttl;
  nat[cpos].tstamp  = timestamp;
  nat[cpos].ip      = ip->saddr;
  if (!cfg.notresolv)
    strncpy(nat[cpos].hostname, resolv(ip->saddr), sizeof(nat[cpos].hostname));
  nat[cpos].time    = cur_time;
  nat[cpos].prob    = 0;

  nat[cpos].oscnt   = 1;
  strncpy(nat[cpos].osgenre, osgenre, sizeof(nat[cpos].osgenre));

  bzero(nat[cpos].osl, sizeof(nat[cpos].osl));
  strncpy(nat[cpos].osl[0], osgenre, sizeof(nat[cpos].osl[0]));

  nat[cpos].synlencnt = 1;
  bzero(nat[cpos].synlen, sizeof(nat[cpos].synlen));
  nat[cpos].synlen[0] = ip->tot_len;

  cpos++;

end: ;

}


static void usage(void) {
  fprintf(stderr,
#ifndef WIN32
  "Usage: natdet [-Idpv] [-i DEV] [-l FILE] [-m PROB] [-f FACTORS] ['filter expr']\n\n"
#else
  "Usage: natdet [-Wpv] [-i DEV] [-l FILE] [-m PROB] [-f FACTORS] [\"filter expr\"]\n\n"
#endif
  "  -l FILE       - dump output to this logfile\n"
  "  -i DEV        - use this interface\n"
  "  -m PROB       - minimum probability level in percent (default = 50)\n"
  "  -f FACTORS    - enable only specified factors with colon as separator\n"
  "                  available: OSGENRE:TSTAMP:TTL:SRCPORT:MSSWSS:SYNLEN\n"
#ifndef WIN32
  "  -I            - switch to \"I-hate-logs!\" mode\n"
  "  -d            - run in background\n"
#else
  "  -W            - list available devices\n"
#endif
  "  -p            - open interface in PROMISC mode\n"
  "  -n            - do NOT convert IP addresses to names\n"
  "  -v            - verbose output (useful)\n"
  "  -vv           - more verbose (for debugging)\n\n"
#ifndef WIN32
  "Example: natdet -v -l spidey.log -i eth0 -d\n\n");
#else
  "Example: natdet -v -l spidey.log -i 1\n\n");
#endif

  exit(0);
}


#ifndef WIN32
static void int_sig() {
  fprintf(stderr, "[-] caught SIGINT(2) signal, exiting...\n");
  if (cfg.iactive) {
    curs_set(1);
    endwin();
  }
  exit(2);
}


static void term_sig() {
  fprintf(stderr, "[-] caught SIGTERM(15) signal, exiting...\n");
  if (cfg.iactive) {
    curs_set(1);
    endwin();
  }
  exit(15);
}


static void segv_sig() {
  fprintf(stderr, "[!] Damn, SIGSEGV signal (segmentation fault)!\n");
  if (cfg.iactive) {
    curs_set(1);
    endwin();
  }
  exit(11);
}
#endif


int main(int argc, char *argv[]) {
#ifdef WIN32
  pcap_if_t *alldevs, *d;
  int adapter;
#endif
  char errbuff[PCAP_ERRBUF_SIZE],
       buff[256];
  u_char i;

  struct bpf_program filter;
  struct passwd *nobody;

  pcap_t *pt;
#ifndef WIN32
  pid_t  pid;

  signal(SIGSEGV, &segv_sig);
  signal(SIGINT, &int_sig);
  signal(SIGTERM, &term_sig);
#endif

  cfg.minprob = MIN_PROB;
  cfg.factor = OSGENRE|TSTAMP|TTL|SRCPORT|MSSWSS|SYNLEN;

  for (i = 1; i < argc; i++) {

    if (!strcmp("-v", argv[i])) {
      cfg.verbose = 1;

    } else if (!strcmp("-vv", argv[i])) {
      cfg.verbose = 2;

    } else if (!strcmp("-p", argv[i])) {
      cfg.promisc = 1;

    } else if (!strcmp("-n", argv[i])) {
      cfg.notresolv = 1;

#ifndef WIN32
    } else if (!strcmp("-d", argv[i])) {
      cfg.detach  = 1;

    } else if (!strcmp("-I", argv[i])) {
      cfg.iactive = 1;
      if (cfg.minprob == MIN_PROB)
        cfg.minprob = 1;
#else
    } else if (!strcmp("-W", argv[i])) {
      if (pcap_findalldevs(&alldevs, errbuff) == -1)
        bug("pcap_findalldevs: %s\n", errbuff);
      for (i = 1, d = alldevs; d ; d = d->next, i++) {
        printf("%d. %s ", i, d->name);
        if (d->description) printf(" (%s)\n", d->description);
        else printf(" (No description available)\n");
      }
      pcap_freealldevs(alldevs);
      exit(0);
#endif

    } else if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      usage();

    } else if (!strcmp("-m", argv[i]) && i < argc - 1) {
      i++;
      if (atoi(argv[i]) >= 1 && atoi(argv[i]) <= 100) {
        cfg.minprob = atoi(argv[i]);
      } else {
        bug("invalid probability level: %d\n", atoi(argv[i]));
      }

    } else if (!strcmp("-f", argv[i]) && i < argc - 1) {
      i++;
      cfg.factor = 0;
      char *token = strtok(argv[i], ":");
      while (token != NULL) {
        //printf("[%s]\n", token);
        if (!strcmp(token, "OSGENRE")) cfg.factor |= OSGENRE;
        if (!strcmp(token, "TSTAMP"))  cfg.factor |= TSTAMP;
        if (!strcmp(token, "TTL"))     cfg.factor |= TTL;
        if (!strcmp(token, "SRCPORT")) cfg.factor |= SRCPORT;
        if (!strcmp(token, "MSSWSS"))  cfg.factor |= MSSWSS;
        if (!strcmp(token, "SYNLEN"))  cfg.factor |= SYNLEN;

        token = strtok(NULL, ":");
      }

    } else if (!strcmp("-i", argv[i]) && i < argc - 1) {
      i++;
      iface = argv[i];
      if (strlen(iface) > 64) {
        printf("Nice try ;-)\n"); /* Don't say anything :) */
        exit(0);
      }

    } else if (!strcmp("-l", argv[i]) && i < argc - 1) {
      i++;
      cfg.logfile = 1;
      if (strlen(argv[i]) > 256) {
        printf("Nice try ;-)\n"); /* Don't say anything */
        exit(0);
      }
      if (!freopen(argv[i], "a", stdout))
        bug("unable to open '%s'\n", argv[i]);
      setvbuf(stdout, (char *)NULL, _IONBF, 0);
      printf("NATDet %s started %s\n", VERSION, showtime());

    } else if (argc > 1) {
      if (strcmp(argv[i - 1], "-i") &&
          strcmp(argv[i - 1], "-l") &&
          strcmp(argv[i - 1], "-m")) {
        snprintf(buff, sizeof(buff), "%s and tcp[13] == 2", argv[i]);
        rule = buff;
      }
    }

  }

  if (!rule)
    rule = "tcp[13] == 2";

  if (!cfg.logfile) {
    printf("NATDet "VERSION"\n"
	   "(c) Marcin Ulikowski <elceef@itsec.pl>\n");
  }

#ifndef WIN32
  if (geteuid())
    bug("euid=0 is required (current=%d)\n", geteuid());

  if (!iface)
    iface = pcap_lookupdev(errbuff);
#else
  if (pcap_findalldevs(&alldevs, errbuff) == -1)
    bug("pcap_findalldevs: %s\n", errbuff);

  if (!iface) {
    d = alldevs;
  } else {
    adapter = atoi(iface);
    for (i = 1, d = alldevs; adapter && i < adapter && d; i++, d = d->next);
    if (!d) bug("unable to find device '%d'\n", adapter);
  }

  iface = d->name;
#endif

  if (!(pt = pcap_open_live(iface, PACKET_LEN, cfg.promisc, 1, errbuff)))
    bug("unable to open device '%s'\n", iface);

  // privsep
  nobody = getpwnam("nobody");
  if (nobody && nobody->pw_uid)
    setuid(nobody->pw_uid);
  else
    setuid(getuid());

  set_fhlen(pcap_datalink(pt));

  load_signatures();

  printf("[+] device: '%s', signatures: %d\n", iface, sigs);

  if (strcmp("tcp[13] == 2", rule)) {
    printf("[+] filter: '%s'\n", rule);
  }
#ifndef WIN32
  else {
    u_int netp, maskp, maskb = 0;
    if (!pcap_lookupnet(iface, &netp, &maskp, errbuff)) {
      while (maskp)  {
        maskb++ ;
        maskp &= (maskp - 1) ;
      }
      if (netp)
        printf("[i] suggested filter: 'src net %s/%u'\n", long2ip(netp), maskb);
    }
  }
#endif

  if (pcap_compile(pt, &filter, rule, 1, 0))
    bug("syntax error: %s\n", pcap_geterr(pt));

#ifndef WIN32
  if (cfg.detach && cfg.logfile && !cfg.iactive) {
    pid = fork();
    if (pid < 0) bug("cannot fork()\n");
    if (pid) exit(0);
    dup2(1, 2);
    close(0);
    setsid();
    signal(SIGHUP, SIG_IGN);
    fflush(0);

  } else if (cfg.detach && !cfg.logfile) {
    bug("switch -d requires -l\n");

  } else if (cfg.detach && cfg.iactive) {
    bug("can't use -I mode in background, use screen(1) instead\n");

  } else if (cfg.logfile && cfg.iactive) {
    bug("can't log to file in 'I-hate-logs!' mode\n");
  }
#endif

#ifdef _DEBUG
  printf("[i] Compiled in debug mode!\n");
#endif

  if (!strcmp(iface, "any") && !strcmp(rule, "tcp[13] == 2")) {
    printf("[i] It is recommended to use filter expression with this interface!\n");
  }

  pcap_setfilter(pt, &filter);

  if (cfg.factor) {
    printf("[+] Active factors: ");
    if (cfg.factor & OSGENRE) printf("OSGENRE ");
    if (cfg.factor & TSTAMP)  printf("TSTAMP ");
    if (cfg.factor & TTL)     printf("TTL ");
    if (cfg.factor & SRCPORT) printf("SRCPORT ");
    if (cfg.factor & MSSWSS)  printf("MSSWSS ");
    if (cfg.factor & SYNLEN)  printf("SYNLEN ");
    putchar('\n');
  } else {
    bug("error in factor string, check -f option\n");
  }

#ifndef WIN32
  if (cfg.iactive) {
    initscr();
    curs_set(0);
    printw("Awaiting packets ...\n");
    refresh();
  }
#endif
  if (!cfg.iactive) {
    printf("[+] Awaiting packets ...\n");
  }

  pcap_loop(pt, -1, (pcap_handler)&watch, 0);

  pcap_close(pt);

  return 0;
}
